from CPACqc.main import main
import os
import shutil
from colorama import Fore, Style
import pkg_resources

def run():
    import argparse

    # Parse command-line arguments
    parser = argparse.ArgumentParser(description="Process BIDS directory and generate QC plots.")
    parser.add_argument("-d", "--bids_dir", required=True, help="Path to the BIDS directory")
    parser.add_argument("-o", "--qc_dir", required=True, help="Path to the QC output directory")
    parser.add_argument("-c", "--config", required=False, help="Config file")
    parser.add_argument("-s", "--sub", required=False, help="specify subject/participant label")
    parser.add_argument("-n", "--n_procs", type=int, default=8, help="Number of processes to use for multiprocessing")

    args = parser.parse_args()
    
    try:
        # Create the QC output directory if it doesn't exist
        os.makedirs(args.qc_dir, exist_ok=True)

        # Locate the templates directory within the package
        templates_dir = pkg_resources.resource_filename('bids_qc', 'templates')
        # Copy only the index.html file from the templates directory to the QC output directory
        src_file = os.path.join(templates_dir, 'index.html')
        dest_file = os.path.join(args.qc_dir, 'index.html')
        shutil.copy2(src_file, dest_file)
    except Exception as e:
        print(f"Error copying templates: {e}")
        pass

    not_plotted = main(args.bids_dir, args.qc_dir, args.config, args.sub, args.n_procs)
    if len(not_plotted) > 0:
        print(Fore.RED + "Some files were not plotted. Please check log")
    else:
        print(Fore.GREEN + "All files were successfully plotted")
    print(Style.RESET_ALL)


from setuptools import setup, find_packages

# Read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

# Read the requirements from the requirements.txt file
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='CPACqc',  
    version='0.1.2',  
    author='Biraj Shrestha',  
    author_email='birajstha@gmail.com',  
    description='A package to view Nifti files in a BIDS dataset and generate QC plots.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/birajstha/bids_qc',  
    packages=find_packages(),
    install_requires=requirements,
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',  # Replace with your license
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.10',
    entry_points={
        'console_scripts': [
            'cpacqc=CPACqc.cli:run',  # This points to the run function in cli.py
        ],
    },
    include_package_data=True,
    package_data={
        'CPACqc': ['templates/index.html'],
    },
)


import pandas as pd
import os
from concurrent.futures import ProcessPoolExecutor, as_completed
from tqdm import tqdm

from bids2table import bids2table
import nibabel as nib
from colorama import Fore, Style, init
import logging

from CPACqc.utils import *
from CPACqc.plot import run

def setup_logger(qc_dir):
    # setup logger
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)

    # Create a file handler
    log_file = os.path.join(qc_dir, 'qc.log')
    file_handler = logging.FileHandler(log_file)
    file_handler.setLevel(logging.INFO)

    # Create a logging format
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    file_handler.setFormatter(formatter)

    # Add the handlers to the logger
    logger.addHandler(file_handler)
    
    return logger

def main(bids_dir, qc_dir, config=False, sub=None, n_procs=8):
    os.makedirs(qc_dir, exist_ok=True)
    logger = setup_logger(qc_dir)
    
    logger.info(f"Running QC with nprocs {n_procs}...")
    
    csv_file = os.path.join(qc_dir, "df.csv")

    plots_dir = os.path.join(qc_dir, "plots")
    os.makedirs(plots_dir, exist_ok=True)

    overlay_dir = os.path.join(qc_dir, "overlays")
    os.makedirs(overlay_dir, exist_ok=True)

    df = parse_bids(bids_dir, workers=n_procs, logger=logger)

    if sub:
        if 'sub-' in sub:
            sub = sub.split('-')[-1]
        # Check if sub matches the available subjects
        if sub not in df['sub'].unique():
            logger.error(f"Subject {sub} not found in the DataFrame. Available subjects: {df['sub'].unique()}")
            print(Fore.RED + f"Subject {sub} not found in the DataFrame. Available subjects: {df['sub'].unique()}" + Style.RESET_ALL)
            return False
        
        df = df[df['sub'] == sub]
        
    #df.to_csv(csv_file, index=False)
    
    for col in df.columns:
        if isinstance(df[col].iloc[0], dict):
            df[col] = df[col].apply(lambda x: str(x) if x else "")
            if df[col].nunique() == 1 and df[col].iloc[0] == "":
                df = df.drop(columns=[col])
                
    # give me all columns that have more than one unique value and drop other columns
    # non_single_value_columns = df.columns[df.nunique() > 1].tolist()
    # df = df[non_single_value_columns]

    # fill all columns with NaN with empty string
    df = df.fillna("")

    # drop json column too
    df = df.drop(columns=["json"])

    # give me all whose ext is nii.gz
    nii_gz_files = df[df.file_path.str.endswith(".nii.gz")].copy()

    # add one column that breaks the file_path to the last name of the file and drops extension
    nii_gz_files.loc[:, "file_name"] = nii_gz_files.file_path.apply(lambda x: os.path.basename(x).replace(".nii.gz", ""))

    nii_gz_files.loc[:, "resource_name"] = nii_gz_files.apply(gen_resource_name, axis=1)

    nii_gz_files = nii_gz_files[nii_gz_files.file_path.apply(lambda x: is_3d_or_4d(x, logger))]

    # save nii_gz_files to csv
    nii_gz_files_csv_path = os.path.join(qc_dir, "nii_gz_files.csv")
    nii_gz_files.to_csv(nii_gz_files_csv_path, index=False)

    # for rows in overlay_csv find the resource_name and get the rows
    if config:
        overlay_df = pd.read_csv(overlay_csv)
        overlay_df = overlay_df.fillna(False)
        results = overlay_df.apply(lambda row: process_row(row, nii_gz_files, overlay_dir, plots_dir, logger), axis=1).tolist()

        # Flatten the list of lists
        results = [item for sublist in results for item in sublist]

        # Create a DataFrame from the results
        result_df = pd.DataFrame(results)
    else:
        result_df = nii_gz_files.copy()
        result_df['file_path_1'] = nii_gz_files['file_path']
        result_df['file_path_2'] = None
        result_df['file_name'] = result_df.apply(lambda row: gen_filename(res1_row=row), axis=1)
        result_df['plots_dir'] = plots_dir
        result_df['plot_path'] = result_df.apply(lambda row: generate_plot_path(create_directory(row['sub'], row['ses'], row['plots_dir']), row['file_name']), axis=1)
        
        columns_to_keep = ['sub', 'ses', 'file_path_1', 'file_path_2', 'file_name', 'plots_dir', 'plot_path']
        result_df = result_df[columns_to_keep].copy()

    result_df['relative_path'] = result_df.apply(lambda row: os.path.relpath(row['plot_path'], qc_dir), axis=1)
    
    # save the result_df to csv
    result_df_csv_path = os.path.join(qc_dir, "results.csv")
    if os.path.exists(result_df_csv_path):
        result_df.to_csv(result_df_csv_path, mode='a', header=False, index=False)
    else:
        result_df.to_csv(result_df_csv_path, index=False)
    
    args = [
        (
            row['sub'], 
            row['ses'],  
            row['file_path_1'],
            row['file_path_2'], 
            row['file_name'],
            row['plots_dir'],
            row['plot_path'],
            logger
        ) 
        for _, row in result_df.iterrows()
    ]

    not_plotted = []
    # Use concurrent.futures to process each row with the specified number of processes
    with ProcessPoolExecutor(max_workers=n_procs) as executor:
        futures = {executor.submit(run_wrapper, arg): arg for arg in args}
        for future in tqdm(as_completed(futures), total=len(futures), desc="Processing ..."):
            try:
                future.result()
            except Exception as e:
                if "terminated abruptly" in str(e):
                    print(Fore.RED + f"Error processing {futures[future]}: {e}\n Try with lower number of processes" + Style.RESET_ALL)
                logger.error(f"Error processing {futures[future]}: {e}, Try with a lower number of processes")
                not_plotted.append(futures[future])
    return not_plotted
    
if __name__ == "__main__":
    import argparse

    # Parse command-line arguments
    parser = argparse.ArgumentParser(description="Process BIDS directory and generate QC plots.")
    parser.add_argument("--bids_dir", required=True, help="Path to the BIDS directory")
    parser.add_argument("--qc_dir", required=True, help="Path to the QC output directory")
    parser.add_argument("--config", required=False, help="Config file")
    parser.add_argument("--n_procs", type=int, default=10, help="Number of processes to use for multiprocessing")

    args = parser.parse_args()
    main(args.bids_dir, args.qc_dir, args.config, args.n_procs)


import pandas as pd
import os
import argparse
from multiprocessing import Pool
from tqdm import tqdm
from functools import lru_cache
import tempfile

from bids2table import bids2table
import nibabel as nib
from colorama import Fore, Style, init

from CPACqc.plot import run

import re

def gen_resource_name(row):
    sub = row["sub"]
    ses = row["ses"] if row["ses"] != "" else False

    sub_ses = f"sub-{sub}_ses-{ses}" if ses else f"sub-{sub}"

    task = row["task"] if row["task"] != "" else False
    run = row["run"] if row["run"] != "" else False
    
    # Create a flexible pattern for the scan part
    scan = f"task-{task}_run-\\d*_" if task and run else ""
    
    # Use regular expression to replace the pattern
    pattern = re.escape(f"{sub_ses}_") + scan
    resource_name = re.sub(pattern, "", row["file_name"])
    
    return resource_name

# add a utility function to return rows provided a resource_name
def get_rows_by_resource_name(resource_name, nii_gz_files, logger):
    # get all rows that have the resource_name
    rows = nii_gz_files[nii_gz_files.resource_name == resource_name]
    if len(rows) == 0:
        logger.error(f"NOT FOUND: {resource_name}")
        return None
    return rows

# check file_path and drop the ones that are higher dimensions for now
def is_3d_or_4d(file_path, logger):
    dim = len(nib.load(file_path).shape)
    if dim > 4:
        file_name = os.path.basename(file_path).split(".")[0]
        logger.error(f"NOT 3D: {file_name} \n its {dim}D")
        logger.error(f"Skipping for now ....")
        return False
    return True

def gen_filename(res1_row, res2_row=None):
    scan = f"task-{res1_row['task']}_run-{int(res1_row['run'])}_" if res1_row['task'] and res1_row['run'] else ""
    if res2_row is not None:
        return f"sub-{res1_row['sub']}_ses-{res1_row['ses']}_{scan + res1_row['resource_name']} overlaid on {res2_row['resource_name']}"
    else:
        return f"sub-{res1_row['sub']}_ses-{res1_row['ses']}_{scan + res1_row['resource_name']}"

def create_directory(sub, ses, base_dir):
    sub_dir = os.path.join(base_dir, sub, ses)
    os.makedirs(sub_dir, exist_ok=True)
    return sub_dir

def generate_plot_path(sub_dir, file_name):
    return os.path.join(sub_dir, f"{file_name}.png")

def process_row(row, nii_gz_files, overlay_dir, plots_dir, logger):
    image_1 = row.get("image_1", False)
    image_2 = row.get("image_2", False)

    resource_name_1 = get_rows_by_resource_name(image_1, nii_gz_files, logger) if image_1 else None
    resource_name_2 = get_rows_by_resource_name(image_2, nii_gz_files, logger) if image_2 else None

    if resource_name_1 is None:
        logger.error(f"NOT FOUND: {image_1}")
        return []

    result_rows = []
    seen = set()  # To track duplicates

    for _, res1_row in resource_name_1.iterrows():
        if resource_name_2 is not None:
            for _, res2_row in resource_name_2.iterrows():
                file_name = gen_filename(res1_row, res2_row)
                if file_name not in seen:
                    seen.add(file_name)
                    sub_dir = create_directory(res1_row['sub'], res1_row['ses'], overlay_dir)
                    plot_path = generate_plot_path(sub_dir, file_name)
                    result_rows.append({
                        "sub": res1_row["sub"],
                        "ses": res1_row["ses"],
                        "file_path_1": res1_row["file_path"],
                        "file_path_2": res2_row["file_path"],
                        "file_name": file_name,
                        "plots_dir": overlay_dir,
                        "plot_path": plot_path
                    })
        else:
            file_name = gen_filename(res1_row)
            if file_name not in seen:
                seen.add(file_name)
                sub_dir = create_directory(res1_row['sub'], res1_row['ses'], plots_dir)
                plot_path = generate_plot_path(sub_dir, file_name)
                result_rows.append({
                    "sub": res1_row["sub"],
                    "ses": res1_row["ses"],
                    "file_path_1": res1_row["file_path"],
                    "file_path_2": None,
                    "file_name": file_name,
                    "plots_dir": plots_dir,
                    "plot_path": plot_path
                })

    return result_rows

@lru_cache(maxsize=None)
def parse_bids(base_dir, workers=8, logger=None):
    print(Fore.YELLOW + "Parsing BIDS directory..." + Style.RESET_ALL)
    if logger: 
        logger.info("Parsing BIDS directory...")
    df = bids2table(base_dir, workers=workers).flat
    return df

def run_wrapper(args):
    return run(*args)


import pandas as pd
from multiprocessing import Pool
import os
from tqdm import tqdm
from colorama import Fore, Style, init
import nibabel as nib
from nilearn.plotting import plot_stat_map
import matplotlib.pyplot as plt
import numpy as np

def plot_nii_overlay(in_nii, plot_loc, background=None, volume=None, cmap='viridis', title=None, alpha=0.8, threshold=20):
    im = nib.load(in_nii)
    if volume is not None:
        v = volume
        im = nib.Nifti1Image(im.get_fdata()[:,:,:,v], header=im.header, affine=im.affine)

    if threshold != 'auto':
        lb = np.percentile(np.sort(np.unique(im.get_fdata())), float(threshold))
    else:
        lb = threshold

    if background and background != 'None':
        bg = nib.load(background)
        plot_stat_map(im, bg_img=bg, output_file=plot_loc,
                      black_bg=True, threshold=lb, title=title, cmap=cmap, alpha=float(alpha))
    
    elif background == None or background == 'None':
        plot_stat_map(im, bg_img=None, output_file=plot_loc,
                      black_bg=True, threshold=lb, title=title, cmap=cmap, alpha=float(alpha))
                                                                                                        
    else:
        plot_stat_map(im, output_file=plot_loc,
                      black_bg=True, threshold=lb, title=title, cmap=cmap, alpha=float(alpha))


def run(sub, ses, file_path_1, file_path_2, file_name, plots_dir, plot_path, logger):
    # # check if the above files exist
    # if not os.path.exists(file_path_1):
    #     print(Fore.RED + f"NO FILE: {file_name}" + Style.RESET_ALL)
    #     return

    # # # Check if the plot already exists
    # if os.path.exists(plot_path):
    #     print(Fore.YELLOW + f"Plot already exists: {file_name}" + Style.RESET_ALL)
    #     return

    # Check dimension and set volume index if 4D
    dim = len(nib.load(file_path_1).shape)
    volume_index = 0 if dim == 4 else None

    try:
        plot_nii_overlay(
            file_path_1,
            plot_path,
            background=file_path_2 if file_path_2 else None,
            volume=volume_index,
            cmap='bwr',
            title="",
            alpha=0.5,
            threshold="auto"
        )
    except Exception as e:
        # print(Fore.RED + f"Error on {file_name}" + Style.RESET_ALL)
        # print(Fore.RED + f"Error: {e}" + Style.RESET_ALL)
        return f"Error on {file_name}: {e}"
    return f"Successfully plotted"